@@ -1,7 +1,6 @@
 {-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE ExistentialQuantification #-}
 {-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE IncoherentInstances #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeFamilies #-}
@@ -29,7 +28,7 @@ module Shelly
        (
          -- * Entering Sh
          Sh, ShIO, shelly, shellyNoDir, shellyFailDir, asyncSh, sub
-         , silently, verbosely, escaping, print_stdout, print_stderr, print_commands
+         , silently, verbosely, escaping, print_stdout, print_stderr, print_commands, print_commands_with
          , onCommandHandles
          , tracing, errExit
          , log_stdout_with, log_stderr_with
@@ -57,7 +56,7 @@ module Shelly
          , cd, chdir, chdir_p, pwd
 
          -- * Printing
-         , echo, echo_n, echo_err, echo_n_err, inspect, inspect_err
+         , echo, echo_n, echo_err, echo_n_err, echoWith, inspect, inspect_err
          , tag, trace, show_command
 
          -- * Querying filesystem
@@ -106,12 +105,12 @@ import Control.Applicative
 import Control.Concurrent
 import Control.Concurrent.Async (async, wait, Async)
 import Control.Exception
-import Control.Monad ( when, unless, void, forM, filterM, liftM2 )
+import Control.Monad ( when, unless, void, liftM2 )
 import Control.Monad.Trans ( MonadIO )
 import Control.Monad.Reader (ask)
 
 import Data.ByteString ( ByteString )
-import Data.Char       ( isAlphaNum, isDigit, isSpace )
+import Data.Char       ( isAlphaNum, isDigit, isSpace, isPrint )
 #if defined(mingw32_HOST_OS)
 import Data.Char       ( toLower )
 #endif
@@ -122,13 +121,11 @@ import Data.Maybe
 import Data.Semigroup  ( (<>) )
 #endif
 import Data.Sequence   ( Seq, (|>) )
-import Data.Set        ( Set )
 import Data.Time.Clock ( getCurrentTime, diffUTCTime  )
 import Data.Tree       ( Tree(..) )
 import Data.Typeable
 
 import qualified Data.ByteString as BS
-import qualified Data.List as List
 import qualified Data.Set as Set
 import qualified Data.Text as T
 import qualified Data.Text.IO as TIO
@@ -137,7 +134,7 @@ import qualified Data.Text.Encoding.Error as TE
 
 import System.Directory
   ( setPermissions, getPermissions, Permissions(..), getTemporaryDirectory, pathIsSymbolicLink
-  , copyFile, removeFile, doesFileExist, doesDirectoryExist, listDirectory
+  , copyFile, removeFile, doesFileExist, doesDirectoryExist
   , renameFile, renameDirectory, removeDirectoryRecursive, createDirectoryIfMissing
   , getCurrentDirectory
   )
@@ -153,34 +150,20 @@ import System.Process
   , ProcessHandle, StdStream(..)
   )
 
-{- GHC won't default to Text with this, even with extensions!
- - see: http://hackage.haskell.org/trac/ghc/ticket/6030
-class CmdArgs a where
-  toTextArgs :: a -> [Text]
-
-instance CmdArgs Text       where toTextArgs t = [t]
-instance CmdArgs FilePath   where toTextArgs t = [toTextIgnore t]
-instance CmdArgs [Text]     where toTextArgs = id
-instance CmdArgs [FilePath] where toTextArgs = map toTextIgnore
-
-instance CmdArgs (Text, Text) where
-  toTextArgs (t1,t2) = [t1, t2]
-instance CmdArgs (FilePath, FilePath) where
-  toTextArgs (fp1,fp2) = [toTextIgnore fp1, toTextIgnore fp2]
-instance CmdArgs (Text, FilePath) where
-  toTextArgs (t1, fp1) = [t1, toTextIgnore fp1]
-instance CmdArgs (FilePath, Text) where
-  toTextArgs (fp1,t1) = [toTextIgnore fp1, t1]
-
-cmd :: (CmdArgs args) => FilePath -> args -> Sh Text
-cmd fp args = run fp $ toTextArgs args
--}
-
 -- | Argument converter for the variadic argument version of 'run' called 'cmd'.
 -- Useful for a type signature of a function that uses 'cmd'.
-class CmdArg a where toTextArg :: a -> Text
-instance CmdArg Text     where toTextArg = id
-instance CmdArg String   where toTextArg = T.pack
+class CmdArg a where
+  -- | @since 1.12.0
+  toTextArgs :: a -> [Text]
+
+instance CmdArg Text   where
+  toTextArgs = (: [])
+
+instance CmdArg String where
+  toTextArgs = (: []) . T.pack
+
+instance {-# OVERLAPPABLE #-} CmdArg a => CmdArg [a] where
+  toTextArgs = concatMap toTextArgs
 
 -- | For the variadic function 'cmd'.
 --
@@ -188,23 +171,17 @@ instance CmdArg String   where toTextArg = T.pack
 class ShellCmd t where
     cmdAll :: FilePath -> [Text] -> t
 
+-- This is the only candidate for `_ <- cmd path x y z` so marking it incoherent will return it and
+-- terminate the search immediately.  This also removes the warning for do { cmd path x y z ; .. }
+-- as GHC will infer `Sh ()` instead of `Sh Text` as before.
+instance {-# INCOHERENT #-} s ~ () => ShellCmd (Sh s) where
+    cmdAll = run_
+
 instance ShellCmd (Sh Text) where
     cmdAll = run
 
-instance (s ~ Text, Show s) => ShellCmd (Sh s) where
-    cmdAll = run
-
--- note that Sh () actually doesn't work for its case (_<- cmd) when there is no type signature
-instance ShellCmd (Sh ()) where
-    cmdAll = run_
-
 instance (CmdArg arg, ShellCmd result) => ShellCmd (arg -> result) where
-    cmdAll fp acc x = cmdAll fp (acc ++ [toTextArg x])
-
-instance (CmdArg arg, ShellCmd result) => ShellCmd ([arg] -> result) where
-    cmdAll fp acc x = cmdAll fp (acc ++ map toTextArg x)
-
-
+    cmdAll fp acc x = cmdAll fp (acc ++ toTextArgs x)
 
 -- | Variadic argument version of 'run'.
 -- Please see the documenation for 'run'.
@@ -605,14 +582,32 @@ whichEith originalFp = whichFull
     whichFull fp = do
       (trace . mappend "which " . toTextIgnore) fp >> whichUntraced
       where
-        whichUntraced | isAbsolute fp          = checkFile
-                      | dotSlash splitOnDirs   = checkFile
-                      | length splitOnDirs > 0 = lookupPath  >>= leftPathError
-                      | otherwise              = lookupCache >>= leftPathError
+        whichUntraced | isAbsolute fp             = checkFile
+                      | startsWithDot splitOnDirs = checkFile
+                      | otherwise                 = lookupPath  >>= leftPathError
 
         splitOnDirs = splitDirectories fp
-        dotSlash ("./":_) = True
-        dotSlash _ = False
+
+        -- 'startsWithDot' receives as input the result of 'splitDirectories',
+        -- which will include the dot (\".\") as its first element only if this
+        -- is a path of the form \"./foo/bar/baz.sh\". Check for example:
+        --
+        -- > import System.FilePath as FP
+        -- > FP.splitDirectories "./test/data/hello.sh"
+        -- [".","test","data","hello.sh"]
+        -- > FP.splitDirectories ".hello.sh"
+        -- [".hello.sh"]
+        -- > FP.splitDirectories ".test/hello.sh"
+        -- [".test","hello.sh"]
+        -- > FP.splitDirectories ".foo"
+        -- [".foo"]
+        --
+        -- Note that earlier versions of Shelly used
+        -- \"system-filepath\" which also has a 'splitDirectories'
+        -- function, but it returns \"./\" as its first argument,
+        -- so we pattern match on both for backward-compatibility.
+        startsWithDot (".":_)  = True
+        startsWithDot _ = False
 
         checkFile :: Sh (Either String FilePath)
         checkFile = do
@@ -637,33 +632,8 @@ whichEith originalFp = whichFull
             res <- liftIO $ isExecutable fullFp
             return $ if res then Just fullFp else Nothing
 
-        lookupCache :: Sh (Maybe FilePath)
-        lookupCache = do
-            pathExecutables <- cachedPathExecutables
-            return $ fmap (flip (</>) fp . fst) $
-                List.find (Set.member fp . snd) pathExecutables
-
-
         pathDirs = mapM absPath =<< ((map T.unpack . filter (not . T.null) . T.split (== searchPathSeparator)) `fmap` get_env_text "PATH")
 
-        cachedPathExecutables :: Sh [(FilePath, Set FilePath)]
-        cachedPathExecutables = do
-          mPathExecutables <- gets sPathExecutables
-          case mPathExecutables of
-              Just pExecutables -> return pExecutables
-              Nothing -> do
-                dirs <- pathDirs
-                executables <- forM dirs (\dir -> do
-                    files <- (liftIO . listDirectory) dir `catch_sh` (\(_ :: IOError) -> return [])
-                    exes <- fmap (map snd) $ liftIO $ filterM (isExecutable . fst) $
-                      map (\f -> (f, takeFileName f)) files
-                    return $ Set.fromList exes
-                  )
-                let cachedExecutables = zip dirs executables
-                modify $ \x -> x { sPathExecutables = Just cachedExecutables }
-                return $ cachedExecutables
-
-
 -- | A monadic findMap, taken from MissingM package
 findMapM :: Monad m => (a -> m (Maybe b)) -> [a] -> m (Maybe b)
 findMapM _ [] = return Nothing
@@ -881,6 +851,13 @@ print_stderr shouldPrint a =
 print_commands :: Bool -> Sh a -> Sh a
 print_commands shouldPrint a = sub $ modify (\st -> st { sPrintCommands = shouldPrint }) >> a
 
+
+-- | Create a sub-Sh in which commands are sent to the user-defined function.
+--
+-- @since 1.12.1
+print_commands_with :: (Text -> IO ()) -> Sh a -> Sh a
+print_commands_with fn a = sub $ modify (\st -> st { sPrintCommandsFn = fn }) >> a
+
 -- | Enter a sub-Sh that inherits the environment
 -- The original state will be restored when the sub-Sh completes.
 -- Exceptions are propagated normally.
@@ -984,6 +961,7 @@ shelly' ros action = do
                    , sPrintStdout = True
                    , sPrintStderr = True
                    , sPrintCommands = False
+                   , sPrintCommandsFn = TIO.hPutStrLn stdout
                    , sInitCommandHandles = initAllHandles (const $ return ())
                    , sCommandEscaping = True
                    , sEnvironment = environment
@@ -1056,11 +1034,23 @@ instance Exception RunFailed
 
 show_command :: FilePath -> [Text] -> Text
 show_command exe args =
-    T.intercalate " " $ map quote (toTextIgnore exe : args)
-  where
-    quote t | T.any (== '\'') t = t
-    quote t | T.any isSpace t = surround '\'' t
-    quote t | otherwise = t
+  let escape char | char `Set.member` specialsInQuotes = T.pack ['\\', char]
+      escape char = T.singleton char
+      quote arg = surround '"' $ T.concatMap escape arg
+      isSafe c = all ($ c) [isPrint, not . isSpace, (`Set.notMember` specials)]
+      showArg "" = surround '"' ""
+      showArg arg | T.all isSafe arg = arg
+      showArg arg = quote arg
+  in T.intercalate " " $ map showArg (toTextIgnore exe : args)
+
+-- | Characters that need to be escaped or quoted to retain their literal value.
+specials :: Set.Set Char
+specials = Set.fromList "\\'\"`$&|;(){}<>"
+
+-- | When inside quotes, characters that need to be escaped to retain their
+-- literal value.
+specialsInQuotes :: Set.Set Char
+specialsInQuotes = Set.fromList "\\\"`$"
 
 -- quote one argument
 quoteOne :: Text -> Text
@@ -1284,7 +1274,7 @@ runHandles exe args reusedHandles withHandles = do
     state <- get
 
     let cmdString = show_command exe args
-    when (sPrintCommands state) $ echo cmdString
+    when (sPrintCommands state) $ echoWith (sPrintCommandsFn state) cmdString
     trace cmdString
 
     let doRun = if sCommandEscaping state then runCommand else runCommandNoEscape
