@@ -9,9 +9,7 @@
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE InstanceSigs#-}
--- | I started exposing multiple module (starting with one for finding)
--- Base prevented circular dependencies
--- However, Shelly went back to exposing a single module
+
 module Shelly.Base
   (
     Sh(..), ShIO, runSh, State(..), ReadOnlyState(..), StdHandle(..),
@@ -22,7 +20,7 @@ module Shelly.Base
     unpack, gets, get, modify, trace,
     ls, lsRelAbs,
     toTextIgnore,
-    echo, echo_n, echo_err, echo_n_err, inspect, inspect_err,
+    echo, echo_n, echo_err, echo_n_err, echoWith, inspect, inspect_err,
     catchany,
     liftIO, (>=>),
     eitherRelativeTo, relativeTo, maybeRelativeTo,
@@ -31,30 +29,24 @@ module Shelly.Base
     , addTrailingSlash
   ) where
 
-#if defined(__GLASGOW_HASKELL__) && __GLASGOW_HASKELL__ < 706
-import Prelude hiding (FilePath, catch)
-#else
-import Prelude hiding (FilePath)
-#endif
-
 import Data.Text (Text)
-import System.Process( ProcessHandle, StdStream(..) )
+import System.Process( StdStream(..) )
 import System.IO ( Handle, hFlush, stderr, stdout )
 
-import Control.Monad (when, (>=>),
-         liftM
-       )
+import Control.Monad ( when, (>=>) )
+#if !MIN_VERSION_base(4,13,0)
 import Control.Monad.Fail (MonadFail)
+import Control.Applicative (Applicative, (<$>))
+import Data.Monoid (mappend)
+#endif
 import Control.Monad.Base
 import Control.Monad.Trans.Control
-import Control.Applicative (Applicative, (<$>))
 import System.Directory( doesDirectoryExist, listDirectory)
 import System.PosixCompat.Files( getSymbolicLinkStatus, isSymbolicLink )
-import System.FilePath  ( FilePath, isRelative)
+import System.FilePath  ( isRelative)
 import qualified System.FilePath as FP
 import qualified System.Directory as FS
 import Data.IORef (readIORef, modifyIORef, IORef)
-import Data.Monoid (mappend)
 import qualified Data.Text as T
 import qualified Data.Text.IO as TIO
 import Control.Exception (SomeException, catch, throwIO, Exception)
@@ -113,6 +105,7 @@ data State = State
    , sPutStderr :: Text -> IO ()   -- ^ by default, hPutStrLn stderr
    , sPrintStderr :: Bool   -- ^ print stderr of command that is executed
    , sPrintCommands :: Bool -- ^ print command that is executed
+   , sPrintCommandsFn :: Text -> IO () -- ^ how to print commands, default is hputStrLn stdout
    , sInitCommandHandles :: StdInit -- ^ initializers for the standard process handles
                                     -- when running a command
    , sCommandEscaping :: Bool -- ^ when running a command, escape shell characters such as '*' rather
@@ -130,10 +123,10 @@ data StdHandle = InHandle StdStream
                | OutHandle StdStream
                | ErrorHandle StdStream
 
--- | Initialize a handle before using it
+-- | Initialize a handle before using it.
 type HandleInitializer = Handle -> IO ()
 
--- | A collection of initializers for the three standard process handles
+-- | A collection of initializers for the three standard process handles.
 data StdInit =
     StdInit {
       inInit :: HandleInitializer,
@@ -141,13 +134,13 @@ data StdInit =
       errInit :: HandleInitializer
     }
 
--- | A monadic-conditional version of the "when" guard.
+-- | A monadic-conditional version of the 'when' guard.
 whenM :: Monad m => m Bool -> m () -> m ()
 whenM c a = c >>= \res -> when res a
 
--- | Makes a relative path relative to the current Sh working directory.
+-- | Makes a relative path relative to the current 'Sh' working directory.
 -- An absolute path is returned as is.
--- To create an absolute path, use 'absPath'
+-- To create an absolute path, use 'absPath'.
 relPath :: FilePath -> Sh FilePath
 relPath fp = do
   wd  <- gets sDirectory
@@ -156,9 +149,10 @@ relPath fp = do
     Right p -> p
     Left  p -> p
 
-eitherRelativeTo :: FilePath -- ^ anchor path, the prefix
-                 -> FilePath -- ^ make this relative to anchor path
-                 -> Sh (Either FilePath FilePath) -- ^ Left is canonic of second path
+eitherRelativeTo
+  :: FilePath                       -- ^ Anchor path, the prefix.
+  -> FilePath                       -- ^ Make this relative to anchor path.
+  -> Sh (Either FilePath FilePath)  -- ^ 'Left' is canonic of second path.
 eitherRelativeTo relativeFP fp = do
   let fullFp = relativeFP FP.</> fp
   let relDir = addTrailingSlash relativeFP
@@ -177,20 +171,20 @@ eitherRelativeTo relativeFP fp = do
       -> Sh (Either FilePath FilePath)
     stripIt rel toStrip nada =
       let stripped = FP.makeRelative rel toStrip
-      in if stripped == toStrip 
-        then nada 
+      in if stripped == toStrip
+        then nada
         else return $ Right stripped
 
--- | make the second path relative to the first
--- Uses 'Filesystem.stripPrefix', but will canonicalize the paths if necessary
-relativeTo :: FilePath -- ^ anchor path, the prefix
-           -> FilePath -- ^ make this relative to anchor path
+-- | Make the second path relative to the first.
+-- Will canonicalize the paths if necessary.
+relativeTo :: FilePath -- ^ Anchor path, the prefix.
+           -> FilePath -- ^ Make this relative to anchor path.
            -> Sh FilePath
 relativeTo relativeFP fp =
   fmap (fromMaybe fp) $ maybeRelativeTo relativeFP fp
 
-maybeRelativeTo :: FilePath -- ^ anchor path, the prefix
-                 -> FilePath -- ^ make this relative to anchor path
+maybeRelativeTo :: FilePath  -- ^ Anchor path, the prefix.
+                 -> FilePath -- ^ Make this relative to anchor path.
                  -> Sh (Maybe FilePath)
 maybeRelativeTo relativeFP fp = do
   epath <- eitherRelativeTo relativeFP fp
@@ -199,23 +193,23 @@ maybeRelativeTo relativeFP fp = do
              Left _ -> Nothing
 
 
--- | add a trailing slash to ensure the path indicates a directory
+-- | Add a trailing slash to ensure the path indicates a directory.
 addTrailingSlash :: FilePath -> FilePath
 addTrailingSlash = FP.addTrailingPathSeparator
 
--- | makes an absolute path.
--- Like 'canonicalize', but on an exception returns 'absPath'
+-- | Make an absolute path.
+-- Like 'canonicalize', but on an exception returns 'absPath'.
 canonic :: FilePath -> Sh FilePath
 canonic fp = do
   p <- absPath fp
   liftIO $ canonicalizePath p `catchany` \_ -> return p
 
 -- | Obtain a (reasonably) canonic file path to a filesystem object. Based on
--- "canonicalizePath" in system-fileio.
+-- 'canonicalizePath'.
 canonicalize :: FilePath -> Sh FilePath
 canonicalize = absPath >=> liftIO . canonicalizePath
 
--- | bugfix older version of canonicalizePath (system-fileio <= 0.3.7) loses trailing slash
+-- | Version of 'FS.canonicalizePath' that keeps a trailing slash.
 canonicalizePath :: FilePath -> IO FilePath
 canonicalizePath p = let was_dir = null (FP.takeFileName p) in
    if not was_dir then FS.canonicalizePath p
@@ -231,12 +225,11 @@ instance Exception EmptyFilePathError
 -- To create a relative path, use 'relPath'.
 absPath :: FilePath -> Sh FilePath
 absPath p | null p = liftIO $ throwIO EmptyFilePathError
-          | isRelative p = do 
+          | isRelative p = do
             cwd <-  gets sDirectory
             return (cwd FP.</> p)
           | otherwise = return p
 
--- | deprecated
 path :: FilePath -> Sh FilePath
 path = absPath
 {-# DEPRECATED path "use absPath, canonic, or relPath instead" #-}
@@ -267,14 +260,14 @@ modify f = do
   state <- ask
   liftIO (modifyIORef state f)
 
--- | internally log what occurred.
+-- | Internally log what occurred.
 -- Log will be re-played on failure.
 trace :: Text -> Sh ()
 trace msg =
   whenM (gets sTracing) $ modify $
     \st -> st { sTrace = sTrace st `mappend` msg `mappend` "\n" }
 
--- | List directory contents. Does *not* include \".\" and \"..\", but it does
+-- | List directory contents. Does /not/ include @.@ and @..@, but it does
 -- include (other) hidden files.
 ls :: FilePath -> Sh [FilePath]
 -- it is important to use path and not absPath so that the listing can remain relative
@@ -289,24 +282,24 @@ lsRelAbs f = absPath f >>= \fp -> do
   let relativized = map (\p -> FP.joinPath [f, p]) files
   return (relativized, absolute)
 
--- | silently uses the Right or Left value of "Filesystem.Path.CurrentOS.toText"
 toTextIgnore :: FilePath -> Text
 toTextIgnore = T.pack
 
--- | a print lifted into 'Sh'
-inspect :: (Show s) => s -> Sh ()
+-- | 'print' lifted into 'Sh'.
+inspect :: Show s => s -> Sh ()
 inspect x = do
-  (trace . T.pack . show) x
-  liftIO $ print x
+  trace $ T.pack s
+  liftIO $ putStrLn s
+  where s = show x
 
--- | a print lifted into 'Sh' using stderr
-inspect_err :: (Show s) => s -> Sh ()
+-- | A 'print' lifted into 'Sh' using stderr.
+inspect_err :: Show s => s -> Sh ()
 inspect_err x = do
   let shown = T.pack $ show x
   trace shown
   echo_err shown
 
--- | Echo text to standard (error, when using _err variants) output. The _n
+-- | Echo text to standard (error, when using @_err@ variants) output. The @_n@
 -- variants do not print a final newline.
 echo, echo_n, echo_err, echo_n_err :: Text -> Sh ()
 echo       msg = traceEcho msg >> liftIO (TIO.putStrLn msg >> hFlush stdout)
@@ -314,6 +307,10 @@ echo_n     msg = traceEcho msg >> liftIO (TIO.putStr msg >> hFlush stdout)
 echo_err   msg = traceEcho msg >> liftIO (TIO.hPutStrLn stderr msg >> hFlush stdout)
 echo_n_err msg = traceEcho msg >> liftIO (TIO.hPutStr stderr msg >> hFlush stderr)
 
+-- | @since 1.12.1
+echoWith :: (Text -> IO ()) -> Text -> Sh ()
+echoWith f msg = traceEcho msg >> liftIO (f msg >> hFlush stdout)
+
 traceEcho :: Text -> Sh ()
 traceEcho msg = trace ("echo " `mappend` "'" `mappend` msg `mappend` "'")
 
@@ -321,4 +318,3 @@ traceEcho msg = trace ("echo " `mappend` "'" `mappend` msg `mappend` "'")
 -- @... `catch` \(e :: SomeException) -> ...@).
 catchany :: IO a -> (SomeException -> IO a) -> IO a
 catchany = catch
-
