@@ -32,13 +32,13 @@ module Shelly.Lifted
          -- please make the same updates here and implements the corresponding
          -- lifted functions.
 
-         -- * Entering Sh.
+         -- * Entering Sh
          Sh, ShIO, S.shelly, S.shellyNoDir, S.shellyFailDir, sub
-         , silently, verbosely, escaping, print_stdout, print_stderr, print_commands
+         , silently, verbosely, escaping, print_stdout, print_stderr, print_commands, print_commands_with
          , tracing, errExit
          , log_stdout_with, log_stderr_with
 
-         -- * Running external commands.
+         -- * Running external commands
          , run, run_, runFoldLines, S.cmd, S.FoldCallback
          , (-|-), lastStderr, setStdin, lastExitCode
          , command, command_, command1, command1_
@@ -50,24 +50,24 @@ module Shelly.Lifted
          , S.StdHandle(..), S.StdStream(..)
 
 
-         -- * Modifying and querying environment.
+         -- * Modifying and querying environment
          , setenv, get_env, get_env_text, get_env_all, appendToPath, prependToPath
 
          -- * Environment directory
          , cd, chdir, chdir_p, pwd
 
          -- * Printing
-         , echo, echo_n, echo_err, echo_n_err, inspect, inspect_err
+         , echo, echo_n, echo_err, echo_n_err, echoWith, inspect, inspect_err
          , tag, trace, S.show_command
 
-         -- * Querying filesystem.
+         -- * Querying filesystem
          , ls, lsT, test_e, test_f, test_d, test_s, test_px, which
 
          -- * Filename helpers
          , absPath, (S.</>), (S.<.>), canonic, canonicalize, relPath, relativeTo
          , S.hasExt
 
-         -- * Manipulating filesystem.
+         -- * Manipulating filesystem
          , mv, rm, rm_f, rm_rf, cp, cp_r, mkdir, mkdir_p, mkdirTree
 
          -- * reading/writing Files
@@ -96,26 +96,23 @@ module Shelly.Lifted
          , followSymlink
          ) where
 
-import qualified Shelly as S
-import Shelly.Base (Sh(..), ShIO, Text, (>=>), FilePath)
+import qualified Shelly      as S
 import qualified Shelly.Base as S
-import Control.Monad ( liftM )
-import Prelude hiding ( FilePath )
+import Shelly.Base     ( Sh(..), ShIO, Text, (>=>) )
+
+import Control.Monad   ( liftM )
 import Data.ByteString ( ByteString )
-import Data.Monoid
-import System.IO ( Handle )
-import Data.Tree ( Tree )
-import qualified System.FilePath as FP
+import Data.Tree       ( Tree )
+import System.IO       ( Handle )
 
 import Control.Exception.Lifted
 import Control.Exception.Enclosed
 import Control.Monad.IO.Class
 import Control.Monad.Trans.Control
 import Control.Monad.Trans.Identity
-import Control.Monad.Trans.List
 import Control.Monad.Trans.Maybe
 import Control.Monad.Trans.Cont
-import Control.Monad.Trans.Error
+import Control.Monad.Trans.Except
 import Control.Monad.Trans.Reader
 import Control.Monad.Trans.State
 import qualified Control.Monad.Trans.State.Strict as Strict
@@ -132,40 +129,46 @@ instance MonadSh Sh where
 
 instance MonadSh m => MonadSh (IdentityT m) where
     liftSh = IdentityT . liftSh
-instance MonadSh m => MonadSh (ListT m) where
-    liftSh m = ListT $ do
-        a <- liftSh m
-        return [a]
+
 instance MonadSh m => MonadSh (MaybeT m) where
     liftSh = MaybeT . liftM Just . liftSh
+
 instance MonadSh m => MonadSh (ContT r m) where
     liftSh m = ContT (liftSh m >>=)
-instance (Error e, MonadSh m) => MonadSh (ErrorT e m) where
-    liftSh m = ErrorT $ do
+
+instance (MonadSh m) => MonadSh (ExceptT e m) where
+    liftSh m = ExceptT $ do
         a <- liftSh m
         return (Right a)
+
 instance MonadSh m => MonadSh (ReaderT r m) where
     liftSh = ReaderT . const . liftSh
+
 instance MonadSh m => MonadSh (StateT s m) where
     liftSh m = StateT $ \s -> do
         a <- liftSh m
         return (a, s)
+
 instance MonadSh m => MonadSh (Strict.StateT s m) where
     liftSh m = Strict.StateT $ \s -> do
         a <- liftSh m
         return (a, s)
+
 instance (Monoid w, MonadSh m) => MonadSh (WriterT w m) where
     liftSh m = WriterT $ do
         a <- liftSh m
         return (a, mempty :: w)
+
 instance (Monoid w, MonadSh m) => MonadSh (Strict.WriterT w m) where
     liftSh m = Strict.WriterT $ do
         a <- liftSh m
         return (a, mempty :: w)
+
 instance (Monoid w, MonadSh m) => MonadSh (RWS.RWST r w s m) where
     liftSh m = RWS.RWST $ \_ s -> do
         a <- liftSh m
         return (a, s, mempty :: w)
+
 instance (Monoid w, MonadSh m) => MonadSh (Strict.RWST r w s m) where
     liftSh m = Strict.RWST $ \_ s -> do
         a <- liftSh m
@@ -181,7 +184,7 @@ instance MonadSh m => S.ShellCmd (m ()) where
     cmdAll = (liftSh .) . S.run_
 
 class Monad m => MonadShControl m where
-    data ShM m a :: *
+    data ShM m a
     liftShWith :: ((forall x. m x -> Sh (ShM m x)) -> Sh a) -> m a
     restoreSh :: ShM m a -> m a
 
@@ -192,15 +195,6 @@ instance MonadShControl Sh where
      {-# INLINE liftShWith #-}
      {-# INLINE restoreSh #-}
 
-instance MonadShControl m => MonadShControl (ListT m) where
-    newtype ShM (ListT m) a = ListTShM (ShM m [a])
-    liftShWith f =
-        ListT $ liftM return $ liftShWith $ \runInSh -> f $ \k ->
-            liftM ListTShM $ runInSh $ runListT k
-    restoreSh (ListTShM m) = ListT . restoreSh $ m
-    {-# INLINE liftShWith #-}
-    {-# INLINE restoreSh #-}
-
 instance MonadShControl m => MonadShControl (MaybeT m) where
     newtype ShM (MaybeT m) a = MaybeTShM (ShM m (Maybe a))
     liftShWith f =
@@ -240,13 +234,12 @@ instance (MonadShControl m, Monoid w)
     {-# INLINE liftShWith #-}
     {-# INLINE restoreSh #-}
 
-instance (MonadShControl m, Error e)
-         => MonadShControl (ErrorT e m) where
-    newtype ShM (ErrorT e m) a = ErrorTShM (ShM m (Either e a))
+instance MonadShControl m => MonadShControl (ExceptT e m) where
+    newtype ShM (ExceptT e m) a = ExceptTShM (ShM m (Either e a))
     liftShWith f =
-        ErrorT $ liftM return $ liftShWith $ \runInSh -> f $ \k ->
-            liftM ErrorTShM $ runInSh $ runErrorT k
-    restoreSh (ErrorTShM m) = ErrorT . restoreSh $ m
+        ExceptT $ liftM return $ liftShWith $ \runInSh -> f $ \k ->
+            liftM ExceptTShM $ runInSh $ runExceptT k
+    restoreSh (ExceptTShM m) = ExceptT . restoreSh $ m
     {-# INLINE liftShWith #-}
     {-# INLINE restoreSh #-}
 
@@ -331,6 +324,10 @@ print_stderr shouldPrint a = controlSh $ \runInSh -> S.print_stderr shouldPrint
 print_commands :: MonadShControl m => Bool -> m a -> m a
 print_commands shouldPrint a = controlSh $ \runInSh -> S.print_commands shouldPrint (runInSh a)
 
+-- | @since 1.12.1
+print_commands_with :: MonadShControl m => (Text -> IO ()) -> m a -> m a
+print_commands_with fn a = controlSh $ \runInSh -> S.print_commands_with fn (runInSh a)
+
 sub :: MonadShControl m => m a -> m a
 sub a = controlSh $ \runInSh -> S.sub (runInSh a)
 
@@ -375,11 +372,11 @@ get = liftSh S.get
 put :: MonadSh m => S.State -> m ()
 put = liftSh . S.put
 
-catch_sh :: (Exception e) => Sh a -> (e -> Sh a) -> Sh a
+catch_sh :: Exception e => Sh a -> (e -> Sh a) -> Sh a
 catch_sh = Control.Exception.Lifted.catch
 {-# DEPRECATED catch_sh "use Control.Exception.Lifted.catch instead" #-}
 
-handle_sh :: (Exception e) => (e -> Sh a) -> Sh a -> Sh a
+handle_sh :: Exception e => (e -> Sh a) -> Sh a -> Sh a
 handle_sh = handle
 {-# DEPRECATED handle_sh "use Control.Exception.Lifted.handle instead" #-}
 
@@ -561,6 +558,10 @@ echo_n     = liftSh . S.echo_n
 echo_err   = liftSh . S.echo_err
 echo_n_err = liftSh . S.echo_n_err
 
+-- | @since 1.12.1
+echoWith :: MonadSh m => (Text -> IO ()) -> Text -> m ()
+echoWith f msg = liftSh $ S.echoWith f msg
+
 relPath :: MonadSh m => FilePath -> m FilePath
 relPath = liftSh . S.relPath
 
@@ -573,7 +574,7 @@ canonic :: MonadSh m => FilePath -> m FilePath
 canonic = liftSh . canonic
 
 -- | Obtain a (reasonably) canonic file path to a filesystem object. Based on
--- "canonicalizePath" in system-fileio.
+-- "canonicalizePath".
 canonicalize :: MonadSh m => FilePath -> m FilePath
 canonicalize = liftSh . S.canonicalize
 
