@@ -4,11 +4,6 @@ module Help (
 ) where
 
 import Shelly
-#if defined(__GLASGOW_HASKELL__) && __GLASGOW_HASKELL__ < 706
-import Prelude hiding ( catch, FilePath )
-#else
-import Prelude hiding ( FilePath )
-#endif
 import Test.HUnit
 import Control.Monad.Trans ( MonadIO )
 
