@@ -7,12 +7,15 @@ import WriteSpec
 import MoveSpec
 import RmSpec
 import FindSpec
+import PrintCommandsFnSpec
 import EnvSpec
 import FailureSpec
 import CopySpec
 import LiftedSpec
 import RunSpec
+import ShowCommandSpec
 import SshSpec
+import PipeSpec
 
 import Test.Hspec
 
@@ -24,9 +27,12 @@ main = hspec $ do
     moveSpec
     rmSpec
     findSpec
+    printCommandsFnSpec
     envSpec
     failureSpec
     copySpec
     liftedSpec
     runSpec
+    showCommandSpec
     sshSpec
+    pipeSpec
